/* Josh Pieper, (c) 2000 */

/* This file is distributed under the GPL, see file COPYING for details */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "gnut_lib.h"
#include "Gnut_List.h"
#include "Gnut_Hash.h"

/*

host history word:

  8 hex digits (32-bit long int)
  defaults to all 0's
  each time a connection is attempted, shifts to the left then OR 0x08
  each time connection opens successfully, OR 0x04
  each time IP address is seen in a packet, low bit set to 1

     A C K I
     ^ ^   ^
     | |   `- IP address gleaned from packets
     | `----- Connection was accepted
     `------- Attempted connection

  K = deliberately killed because of dropped packets or transfer rate

  mask with 0x66666666 to get just the connect history
  host deleted from database when history is nonzero AND history & 0x66666666 is zero

  connect-candidate score = weighted sum of 0x4 bits: last position = 128 points, prev position = 64, and so on. A deliberate kill counts as a failed connection
  when choosing a new host, select highest score in database that isn't already in connection list

I want to figure out why my 'sent' stats are so much lower than
'received'. With 8 connections, keep a 8x8 array of stats showing how
much is being routed through each link of the crossbar, so to speak.
Also extra columns to log how many packets got knocked out pre-routing
and for which reasons (TTL 0, TTL bad, etc.)

*/

static int next_hashstart;

/* CRC-16

The constant 0x1021 is based on the X25 standard CRC polynomial
x^16 + x^12 + x^5 + 1. The highest bit (x616) doesn't need to be
included; the other three bits (12, 5, 1) form the value 0x1021.

The following illustration is from http://www.io.com/~ritter/ARTS/CRCMYST.HTM

       Polynomial = x^5 + x^4 + x^2 + 1 = 110101

       dn   x^4         x^3     x^2         x^1     x^0
       v   +---+       +---+   +---+       +---+   +---+
   +<-XOR<-|Q D|<-XOR<-|Q D|<--|Q D|<-XOR<-|Q D|<--|Q D|<-+
   |       |   |   ^   |   |   |   |   ^   |   |   |   |  |
   |       +---+   |   +---+   +---+   |   +---+   +---+  |
   |               |                   |                  |
   +---------------+-------------------+------------------+

   FIGURE 2.  CRC Hardware

crc = 0xffff;
for each data bit {
  if (high bit of CRC ^ new data bit == 1) {
    crc = (crc << 1) ^ 0x1021;
  } else {
    crc = crc << 1;
  }
}

for CRC-32 (used in e.g. Ethernet) we use 0x04c11db7L, or in binary:
1 0000 0100 1100 0001 0001 1101 1011 0111

There is a long explanation at http://www.snippets.org/CRC.TXT

See also http://cell-relay.indiana.edu/cell-relay/publications/software/CRC/32bitCRC.c.html

If any of these URLs don't work, contact the author Robert Munafo
(www.mrob.com)

*/

#define CRC32_MASK 0x04C11DB7L;

uint32 crc32_conv_8[256];

void crc32_start(uint32 *crc)
{
  int i, j;
  uint32 cctemp;

  *crc = 0xffffffffL;
  for(i=0; i<256; i++) {
  cctemp = ((uint32) i) << 24L;
  for(j=0; j<8; j++) {
    if (cctemp & 0x80000000L) {
    cctemp = (cctemp << 1L) ^ CRC32_MASK;
    } else {
    cctemp <<= 1L;
    }
  }
  crc32_conv_8[i] = cctemp;
  }
}

void crc32_add8(uint32 *crc, uchar data)
{
  uchar index;

  index = (*crc >> 24L) ^ data;
  *crc = (*crc << 8L) ^ crc32_conv_8[index];
}


int gnut_hash_foreach(Gnut_Hash *gh, int (*a)(void *, void *), void *data)
{
  int i, j, k;

  next_hashstart += (rand() & 0xff);
  j = next_hashstart % 256;

  for (i=0;i<256;i++) {
    k = (i + j) % 256;
  /* if (k < 20) { printf("ghfe k=%d\n", k);  } %%%RPM */
    if (gnut_list_foreach(gh->list[k], a, data)==-1) {
      return -1;
    }
  }
  return 0;
}

Gnut_Hash * gnut_hash_new(uchar (*hash_func)(void *),
  int (*compare_func)(void *, void *))
{
  Gnut_Hash *gh;
  int i=0;

  gh=(Gnut_Hash *) xmalloc(sizeof(Gnut_Hash));
  for (i=0;i<256;i++)
  gh->list[i]=NULL;  
  gh->hash_func=hash_func;
  gh->compare_func=compare_func;

  return gh;
}

void gnut_hash_insert(Gnut_Hash *gh, void *data)
{
  int a;

  if (!gh) {
    g_debug(0,"gh=%p\n data=%p\n",gh,data);
    exit(0);
  }

  a=(*gh->hash_func)(data);
  g_debug(5,"hash returned: %i\n",a);

  gh->list[a]=gnut_list_append(gh->list[a],data);
}

void * gnut_hash_find_raw(Gnut_Hash *gh, void *data)
{
  int a;
  int ret;

  Gnut_List *tmp;

  if (!gh) {
    //    g_debug(0,"gh=%p data=%p\n");
    //    exit(0);
    g_debug(5,"not found\n");  
    return NULL;
  }

  for (a = 0; a < 256; a++) {
    g_debug(5,"hash returned: %i\n",a);

    for (tmp=gh->list[a];tmp;tmp=tmp->next) {
      ret=(*gh->compare_func)(data,tmp->data);
      if (ret==0)
    return tmp->data;
    }
  }

  g_debug(5,"not found\n");  
  return NULL;
}

void * gnut_hash_find(Gnut_Hash *gh, void *data)
{
  int a;
  int ret;
  Gnut_List *tmp;

  if (!gh) {
    g_debug(0,"gh=%p data=%p\n",gh,data);
    exit(0);
  }

  a=(*gh->hash_func)(data);

  g_debug(5,"hash returned: %i\n",a);

  for (tmp=gh->list[a];tmp;tmp=tmp->next) {
    gnut_mprobe(tmp);
    ret=(*gh->compare_func)(data,tmp->data);
    if (ret==0)
    return tmp->data;
  }

  g_debug(3,"not found\n");  
  return NULL;
}

uint32 gnut_hash_remove(Gnut_Hash *gh, void *data)
{
  int a;

  if (!gh) {
    g_debug(0,"gh=%p data=%p\n",gh,data);
    exit(0);
  }

  g_debug(3,"entering\n");

  a=(*gh->hash_func)(data);

  gh->list[a]=gnut_list_remove(gh->list[a],data);

  return 0;
}

uint32 gnut_hash_free(Gnut_Hash *gh)
{
  int i;

  g_debug(3,"entering gh=%p\n",gh);

  if (!gh) {
    g_debug(0,"gh=%p\n",gh);
    exit(0);
  }

  for (i=0;i<256;i++)
  gh->list[i]=gnut_list_free(gh->list[i]);

  xfree(gh);
  return 0;
}
