/* Josh Pieper, (c) 2000 */

/* This file is distributed under the GPL, see file COPYING for details */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef WIN32
#include <sys/socket.h>
#include <sys/file.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <netdb.h>
#else
#include <io.h>
#endif

#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <string.h>
#include <pthread.h>
#include <ctype.h>


#include "gnut_lib.h"
#include "gnut_connection.h"
#include "gnut_net.h"
#include "host.h"
#include "gnut_transfer.h"
#include "query.h"
#include "conf.h"
#include "share.h"
#include "cache.h"
#include "gnut_threads.h"
#include "gnut_http.h"

#define SAFE_CLOSE(x) if (x>=0) { shutdown(x,2); close(x); x=-1; }

int incoming_transfer_handle(gnut_transfer *gt, struct sockaddr_in *sin,
  char *request_line);
int gnut_threads_transfer_handle(query_response *qr, gnut_transfer *gt);


int num_outgoing=0;
int num_incoming=0;
int num_downloads=0;
int num_uploads=0;
int count_downloads=0;
int count_uploads=0;

gnutella_packet * current_query_packet;

int gnut_threads_num_downloads()
{
  return num_downloads;
}

int gnut_threads_num_uploads()
{
  return num_uploads;
}

int gnut_threads_count_downloads()
{
  return count_downloads;
}

int gnut_threads_count_uploads()
{
  return count_uploads;
}

int gnut_threads_num_incoming()
{
  return num_incoming;
}

int gnut_threads_num_outgoing()
{
  return num_outgoing;
}

int transfer_cleanup(gnut_transfer *gt, int state)
{
  SAFE_CLOSE(gt->sock);

  gt->state=state;
  
  if (!conf_get_int("auto_clear")) {
    g_debug(2,"waiting for external clear\n");
    while (gt->state==state) sleep(1);
  }
  
  g_debug(2,"done waiting\n");
  
  gnut_transfer_delete(gt);
  return 0;
}

int send_welcome_message(int sock, char *message)
{
  int ret;
  char szBuffer[100];

  g_debug(3,"entering\n");

  sprintf(szBuffer,"GNUTELLA CONNECT/%s\n\n",message);

  do {
    ret = write(sock,szBuffer,strlen(szBuffer));
  } while (errno==EINTR && ret<0);

  if (ret<0) {
    if (gnut_lib_debug>3)
      perror("send_welcome_message");
  }

  return ret;
}

int read_welcome_message(int sock, char *arg)
{
  char buf[100];
  int i;
  int flag=0;

  i=read_line(sock, buf, sizeof(buf));
  g_debug(3, "read_line returned: %i\n", i);

  if (i<=0) return -1;

  if (0) { /* %%%RPM */
    printf("%s rwm: ", arg);
    for(i=0; (i < 100) && (buf[i]); i++) {
      printf("%c", buf[i]);
    }
    printf("\n");
  }

  if (strncmp(buf, "GNUTELLA OK", 11)==0)
    flag=1;
  if (strncmp(buf, "GNUTELLA PASSWORD REQUIRED", 26)==0)
    flag=2;

  if (flag==0)
    return -1;

  read(sock,buf,sizeof(buf));

  return flag;
}

int incoming_gnutella(int sock, struct sockaddr_in *sin)
{
  char *buf;
  int ret;
  gnut_connection *gc;

  g_debug(3,"entering\n");

  /* next we need to read the second line of the welcome */
  buf=(char *) xmalloc(100);
  ret=read_line(sock,buf,100);
  g_debug(3,"read_line returned %i\n",ret);

  if (ret<=0) {
    if (gnut_lib_debug>=1) perror("incoming_gnutella, read_line");
    xfree(buf);
    close(sock);
    return -1;
  }

  /* if authentication is turned on, we must now ask for the 
   * pass phrase */
  if (conf_get_int("password_required")) {
    strcpy(buf,"GNUTELLA PASSWORD REQUIRED\n\n");
    write(sock,buf,strlen(buf));

    // there should be a one line response, consisting of the
    // pass phrase... if it doesn't match, then we abort, otherwise
    // allow the connection
    ret=read_line(sock,buf,100);
    gnut_strstrip(buf);
    g_debug(1,"\"%s\" attempted as a password\n",buf);
    if (ret<=0 || (strcmp(buf,conf_get_str("password"))!=0)) {
      if (gnut_lib_debug>=1)
        perror("incoming_gnutella, read_line2");
      xfree(buf);
      close(sock);
      return -1;
    }
  }

  // ok we've received it alright, now we can send our reply...

  if (++num_incoming>conf_get_int("max_incoming")) {
    if (conf_get_int("redirect")) {
      // drop oldest connection to make room for new
      gnut_connection_kill_oldest();
    } else {
      num_incoming--;
      // we have too many incoming!
      xfree(buf);
      close(sock);
      return 0;
    }
  }

  // first we need to enter ourselves into the connection list
  gc=gnut_connection_new();
  gc->sock=sock;
  gc->ctype=TYPE_INCOMING;
  gc->cstate=STATE_NEGOTIATING;

  memcpy(gc->ip,&sin->sin_addr.s_addr,4);
  gc->port=ntohs(sin->sin_port);

  strcpy(buf,"GNUTELLA OK\n\n");
  ret=write(sock, buf, strlen(buf));

  if (ret<0) {
    perror("incoming_gnutella, write");
    xfree(buf);
    gnut_connection_delete(gc);
    num_incoming--;
    close(sock);
    return -1;
  }

  gc->cstate=STATE_CONNECTED;

  ret = connection_loop(gc);
  g_debug(1, "connection_loop returned: %i\n",ret);

  gc->cstate = STATE_CLEANABLE;

  num_incoming--;

  close(gc->sock);
  xfree(buf);

  gnut_connection_delete(gc);

  g_debug(3,"returning success\n");
  return 0;
  
}

int parse_http_request(char *request, int *ref, char **name)
{
  int iref;
  char *ptr,*ptr2;
  
  g_debug(3,"entering: request=%s\n",request);
  
  ptr=strstr(request,"HTTP/");
  if (ptr==NULL) return -1;
  
  
  g_debug(4,"checked for HTTP\n");
  
  ptr=strstr(request,"/get/");
  if (ptr==NULL) return -2;

  
  g_debug(4,"checked for get\n");
  
  ptr+=5;
  ptr2=strchr(ptr,'/');
  if (ptr2==NULL) return -3;
  
  g_debug(4,"located end slash\n");
  
  *ptr2=0;
  iref=atoi(ptr);
  
  g_debug(4,"located ref num: %i\n",iref);
  
  ptr2++;
  ptr=strstr(ptr2,"HTTP/");
  ptr--;
  for (;ptr>request && (*ptr)==' ';ptr--);
  
  if (ptr==request) return -4;
  
  *(++ptr)=0;
  
  *name=strdup(ptr2);
  *ref=iref;
  
  g_debug(4,"returning success\n");
  return 0;
}

int parse_range(char *buf, int size, int *rangemin, int *rangemax)
{
  int i;
  int num,num2;
  char *ptr;
  
  for (i=6;i<strlen(buf) && !isdigit(buf[i]) && buf[i]!='-';i++);
  
  if (i==strlen(buf)) return 0;
  
  if (buf[i]=='-') {
    num=atoi(&buf[i+1]);
    *rangemin=size-num;
    *rangemax=size;
    return 1;
  }
  
  num=(int) strtol(&buf[i],&ptr,10);
  
  if ((*ptr)!='-') {
    return 0;
  }
  
  ptr++;
  
  if (!isdigit(*ptr)) {
    *rangemin=num;
    *rangemax=size;
    return 1;
  }
  
  num2=(int) strtol(ptr,NULL,10);
  
  *rangemin=num;
  *rangemax=num2;
  return 1;
}

int incoming_transfer_error(int sock)
{
  char a[]="HTTP 404 File Not Found\r\n\r\nThere was an error processing your request\r\n";
  
  write(sock,a,strlen(a));
  return 0;
}

int incoming_transfer(int sock, struct sockaddr_in *sin, char *request_line)
{
  gnut_transfer *gt;

  g_debug(1,"entering num_uploads:%i\n",num_uploads);

  if (++num_uploads>conf_get_int("max_uploads")) {
    num_uploads--;
    incoming_transfer_error(sock);
    write(sock,"The maximum number of uploads at this location has been reached.\r\n",66);
    shutdown(sock,2);
    close(sock);
    return 0;
  }
  
  g_debug(3,"INCOMING TRANSFER!!!\n");
  
  gt=gnut_transfer_new();
  gt->sock=sock;
  gt->type=TYPE_OUTGOING;

  return incoming_transfer_handle(gt,sin,request_line);  
}    

int incoming_transfer_handle(gnut_transfer *gt, struct sockaddr_in *sin,
  char *request_line)
{
  int ret;
  http_request_t *req;

  g_debug(1,"entering\n");
  
  gt->state=STATE_NEGOTIATING;
  memcpy(gt->ip,&sin->sin_addr.s_addr,4);
  gt->port=ntohs(sin->sin_port);
  gt->rate_limit=conf_get_int("default_upload_cap");

  req = http_request_new(gt->sock);
  ret = http_request_parse(req, request_line);
  if (!ret) {
    ret = http_request_read_header(req);
  }
  /*
   * This is not a good request, we want nothing
   * to do with it... shut it down now.
   */
  if (ret) {
    shutdown(gt->sock, 2);

    transfer_cleanup(gt,STATE_ERROR);

    return -1;
  }

  g_debug(1, "Got an HTTP request, processing it\n");

  /*
   * Handle a get request
   */
  if (!strncmp(req->url_file, "/get/",5)) {
    ret = gnut_http_serve_file(gt, req);
    http_request_delete(req);
    num_uploads--;
    if (!ret) count_uploads++;
    transfer_cleanup(gt, (ret==0) ? STATE_DEAD : STATE_ERROR);
    return ret;
  }
  /*
   * Get a directory listing.
   */
#ifdef GNUT_HTTP_SEARCH
  else if ((!strcmp(req->url_file, "/")) && conf_get_int("html_enable")) {
    g_debug(2,"handling file list...\n");
    ret = gnut_http_file_list(gt, req);
    http_request_delete(req);
    num_uploads--;
    
    transfer_cleanup(gt, (ret==0) ? STATE_DEAD : STATE_ERROR);
    return ret;
  }
#endif
  /*
   * Handle a search request.
   */
#ifdef GNUT_HTTP_FILE_LIST
  else if ((!strncmp(req->url_file, "/search/", 8)) && conf_get_int("html_enable")) {
    g_debug(2,"handling search form...\n");
    ret = gnut_http_file_search(gt, req);
    http_request_delete(req);
    num_uploads--;
    transfer_cleanup(gt, (ret==0) ? STATE_DEAD : STATE_ERROR);
    return ret;
  }
#endif

  g_debug(1,"removing http request\n");
  http_request_delete(req);
  num_uploads--;

  transfer_cleanup(gt,STATE_ERROR);
  
  return ret;
}

int incoming_push(int sock, struct sockaddr_in *sin, char *request)
{
  int ret;
  int i;
  char buf[256];
  uchar guid[16];
  uint32 ref;
  uint16 st;
  char *ptr;
  query_response *qr;  
  gnut_transfer *gt;
  
  g_debug(1,"entering sock=%i request=%s\n",sock,request);
  
  
  
  ret=1;
  while (ret>0 || errno==EINTR) {
    g_debug(1,"calling readline\n");
    ret=read_line(sock,buf,256);
    
    g_debug(1,"recvd: %s\n",buf);
    if (strlen(buf)<3) break;
  }
  
  // we need to create the appropriate transfer and query struct,
  // then pass them on to the standard downloader....
  
  ref=atoi(&request[4]);
  ptr=strchr(request,':');
  
  if (ptr==NULL) {
    g_debug(0,"AGGG!\n");
    return 0;
  }
  
  ptr++;
  for (i=0;i<16;i++) {
    strncpy(buf,&ptr[i*2],2);
    buf[2]=0;
    guid[i]=strtol(buf,NULL,16);
  }
  g_debug(2,"ref=%i GUID=%02x%02x%02x%02x...\n",ref,guid[0],guid[1],guid[2],guid[3]);
  g_debug(1,"calling query find...\n");
  
  qr=query_push_find(guid,ref);
  
  if (qr==NULL) {
    g_debug(0,"Woaahhhh!  We don't have this push request in our list!: %s\n",request);
    close(sock);
    return 0;
  }
  
  gt=gnut_transfer_new();
  gt->sock=sock;
  gt->type=TYPE_RECV_PUSH;
  gt->state=STATE_NEGOTIATING;
  gt->total=qr->size;
  memcpy(gt->ip,qr->ip,4);
  memcpy(&st,qr->port,2);
  st=GUINT16_FROM_LE(st);
  gt->port=st;
  gt->rate_limit=conf_get_int("default_download_cap");
  
/*  if (++num_downloads>conf_get_int("max_downloads")) {
    // enter queued state...
    g_debug(1,"I'm getting queued\n");
    num_downloads--;
    gt->cstate=STATE_QUEUED;
    while ((num_downloads>=conf_get_int("max_downloads")) 
      && (gt->cstate==STATE_QUEUED)) sleep(1);
    if (gt->cstate!=STATE_QUEUED) {
      transfer_cleanup(gt,STATE_DEAD);
      xfree(qr->name);
      xfree(qr);
      return 0;
    }
    g_debug(1,"I got  unqueued!\n");
    num_downloads++;
    gt->cstate=STATE_NEGOTIATING;
  }*/
  num_downloads++;
  
  
  ret=gnut_threads_transfer_handle(qr,gt);
  
  g_debug(1,"gnut_threads_transfer_handle returned: %i\n",ret);
  num_downloads--;
  
  return 0;
}

// void *gnut_threads_incoming(void *arg)
//
// main thread for handling incoming connections, socket is
// passed via the pointer, it needs to be free'ed.  First it
// determines whether this is an upload, or gnutella connection,
// then calls the appropriate handler for each...
//
void *gnut_threads_incoming(void *arg)
{
  int sock;
  char buf[256];
  int ret;
  incoming_arg *ia;
    
  g_debug(2,"entering\n");
  
  ia=arg;
  sock=ia->s;  
  
  fcntl(sock,F_SETFL,0);
  
  ret=read_line(sock,buf,sizeof(buf));
  g_debug(1,"read_line returned: %i\n",ret);
  g_debug(1,"the initial line was: %s\n",buf);
  
  if (ret<=0) {
    close(sock);
    xfree(ia);
    return NULL;
  }
  
  if (strncmp(buf,"GNUTELLA CONNECT",16)==0) {
    // this is an incoming gnutella connection...
    incoming_gnutella(sock,&ia->sin);
  } else if (strncmp(buf,"GET ",4)==0) {
    // this is an incoming http connection...
    incoming_transfer(sock,&ia->sin,buf);
  } else if (strncmp(buf,"GIV ",4)==0) {
    // this is a push connection....
    incoming_push(sock,&ia->sin,buf);
  } else {
    close(sock);
  }
  xfree(ia);
  
  g_debug(3,"returning\n");
  return NULL;
}

// this is passed a push_arg struc
// it is called from our program in response to a push packet,
// so it needs to contact the foreign host, do the negotiating,
// then send the file....
void *gnut_threads_push(void *arg)
{
  push_arg *pa;
  gnut_transfer *gt;
  uint16 st;
  char buf[256];
  int ret;
  char *name;
  char *request;
  int i;
  struct sockaddr_in sin;
  
  g_debug(1,"entering\n");
  pa=arg;
  if (++num_uploads > conf_get_int("max_uploads")) {
    num_uploads--;
    xfree(pa->si->path);
    xfree(pa->si->upath);
    xfree(pa->si->fpath);
    xfree(pa->si);
    xfree(pa);
    g_debug(1,"too many uploads!\n");
    
    return NULL;
  }
  
  gt=gnut_transfer_new();
  
  gt->type=TYPE_SEND_PUSH;
  gt->state=STATE_CONNECTING;
  gt->total=pa->si->size;
  gt->fname=strdup(pa->si->path);
  memcpy(gt->ip,pa->ip,4);
  
  memcpy(&st,pa->port,2);
  st=GUINT16_FROM_LE(st);
  
  gt->port=st;
  gt->rate_limit=conf_get_int("default_upload_cap");
  
  g_debug(1,"created new transfer struct\n");
  
  sprintf(buf,"%i.%i.%i.%i",gt->ip[0],gt->ip[1],gt->ip[2],gt->ip[3]);
  
  gt->sock=make_connection(buf, gt->port, gt->ip, 1);
  if (gt->sock<0) {
    g_debug(1,"couldn't connect to %s:%i make_connection returned: %i\n",buf,st,gt->sock);
    close(gt->sock);
    xfree(pa->si->path);
    xfree(pa->si->upath);
    xfree(pa->si->fpath);
    xfree(pa->si);
    xfree(pa);
    num_uploads--;
    
    transfer_cleanup(gt,STATE_ERROR);

    return NULL;
  }
  
  memcpy(&sin.sin_addr.s_addr,gt->ip,4);
  sin.sin_port=gt->port;
  
  g_debug(1,"made connection to %s:%i\n",buf,st);
  
  gt->state=STATE_NEGOTIATING;
  
  request=(char *) xmalloc(4096);
  
  sprintf(request,"GIV %i:",pa->si->ref);
  
  for(i=0;i<16;i++) {
    sprintf(buf,"%02X",(uchar) (conf_get_str("guid"))[i]);
    strcat(request,buf);
  }
  
  name=strrchr(pa->si->path,'/');
  if (name==NULL) name=pa->si->path;
    else name++;
  sprintf(buf,"/%s\n\n",name);
  
  strcat(request,buf);
   
  write(gt->sock,request,strlen(request));
  
  g_debug(1,"connect message sent:\n%s\n",request);
  
  // now we look at the http response and make sure that the
  // correct number of newlines is returned....
  
  while (1) {
    g_debug(1,"trying to read line\n");
    ret=read_line(gt->sock,buf,sizeof(buf));
    if (ret<=0) {
      g_debug(1,"free'ing request\n");
      xfree(request);
      g_debug(1,"closing sock\n");
      close(gt->sock);
      g_debug(1,"free'ing pa->si->path\n"); 
      xfree(pa->si->path);
      g_debug(1,"free'ing pa->si\n");
      xfree(pa->si);
      g_debug(1,"free'ing pa\n");
      xfree(pa);
      num_uploads--;
      
      transfer_cleanup(gt,STATE_ERROR);
      return NULL;
    }
    g_debug(1,"received: %s",buf);
    if (strncmp(buf,"GET ",4)==0) break;
  }

  // now that we've gotten the GET, we can treat this as any normal
  // file upload, so we'll pass it on to the normal upload handler...
  incoming_transfer_handle(gt,&sin,buf);
  
  g_debug(2,"incoming_transfer_handle~ returned, now cleaning up\n");
  
  xfree(request);
  
  g_debug(2,"request free'ed and sock closed\n");
  
  
  xfree(pa->si->path);
  xfree(pa->si->upath);
  xfree(pa->si->fpath);
  xfree(pa->si);
  xfree(pa);
  
  g_debug(2,"push_arg struc free'ed\n");
  
  return NULL;
}    

// this is passed a query_response struc...
void *gnut_threads_transfer(void *arg)
{
  gnut_transfer *gt;
  query_response *qr;
  gnutella_packet *gpa;
  char buf[100];
  uint16 st;
  int ret;
  
  g_debug(2,"starting download procedure...\n");    
  
  qr=arg;

        if (qr->dest_cache) {

          // only one cache download at once
          if (! cache_lock()) {
            free(qr->name);
            free(qr);
            return NULL;
          }

          // empty enough space in the cache
          if (cache_shrink(qr->size) < 0) {
            g_debug(2, "aborting download - couldn't free enough cache space.\n");
            free(qr->name);
            free(qr);
            return NULL;
          }
        }

  
  gt=gnut_transfer_new();
  gt->type=TYPE_INCOMING;
  gt->state=STATE_CONNECTING;
  gt->total=qr->size;
  memcpy(gt->ip,qr->ip,4);
  memcpy(&st,qr->port,2);
  st=GUINT16_FROM_LE(st);
  gt->port=st;
  gt->rate_limit=conf_get_int("default_download_cap");
  gt->fname=strdup(qr->name);
  
  g_debug(3,"created new transfer struc, and set up\n");
  
  
  if (++num_downloads>conf_get_int("max_downloads")) {
    // enter queued state...
    g_debug(1,"I'm queued\n");
    num_downloads--;
    
    if (conf_get_int("verbose")==1)   
      printf("\nDownload of %s queued. (too many downloads)\n",gt->fname);
    gt->state=STATE_QUEUED;
    while ((num_downloads>=conf_get_int("max_downloads")) 
      && (gt->state==STATE_QUEUED)) sleep(1);
    if (gt->state!=STATE_QUEUED) {
      transfer_cleanup(gt,STATE_DEAD);
      free(qr->name);
      free(qr);
      return NULL;
    }
    g_debug(1,"and now I'm not!\n");
    num_downloads++;
    gt->state=STATE_CONNECTING;
  }
  
  sprintf(buf,"%i.%i.%i.%i",qr->ip[0],qr->ip[1],qr->ip[2],qr->ip[3]);
  
  gt->sock=make_connection(buf, st, gt->ip, 1);
  
  if (gt->sock<0) {
    num_downloads--;
      // ignore push cache requests (FIXME - Ray)
        if (qr->dest_cache) {
          cache_unlock();
      transfer_cleanup(gt,STATE_ERROR);   

            free(qr->name);
            free(qr);
            return NULL;
        }
          

    g_debug(1,"couldn't connect to %s:%i make_connection returned: %i\n",buf,st,gt->sock);
    
    transfer_cleanup(gt,STATE_ERROR);
    
    // what I'm going to do now is send out a push request
    // for the file we just couldn't get.... maybe it will
    // work on occasion
    gpa=gp_push_make(conf_get_str("mac"),conf_get_int("ttl"),
      qr->guid,qr->ref,net_local_ip(),
      conf_get_int("local_port"));
    
    query_push_add(qr);  
      
    send_to_all(gpa);
    xfree(gpa->data);
    xfree(gpa);
    
    return NULL;
  }
  gt->port=st;
  
  g_debug(3,"made connection to %s:%i\n",buf,st);

  xfree(gt->fname);
  ret=gnut_threads_transfer_handle(qr,gt);

  num_downloads--;
  return NULL;
}  
  
int gnut_threads_transfer_handle(query_response *qr, gnut_transfer *gt)
{
  char buf[2048];
  int start=0;
  int size=0;
  char *dpath;
  char *request;
  char *shortname;
  int ret;
    char *dest_dir,*c,*c2;
  int cache_download=0;
  struct stat ss;
#if !defined(HAVE_FLOCK) && defined(HAVE_FCNTL)
  struct flock lock;
#endif
  
  if (qr->dest_cache) {
    cache_download=1;
  }
  
  // !!HERE!! is where the entry point needs to be for push connections...
  
  gt->state=STATE_NEGOTIATING;
  
  // first we're going to check to see if the file already exists,
  // if so, then we'll formulate a partial request and use it instead
  shortname=strrchr(qr->name,'/');
    if (shortname==NULL) shortname=qr->name;
      else shortname++;
    dest_dir=(qr->dest_cache ? "cache_path" : "download_path");
    
    if (!conf_get_str(dest_dir)) {
      if (cache_download) {
        g_debug(1,"no cache dir set.\n");
        cache_unlock();
        return -1;
      }
      gt->fname=(char *) xmalloc(strlen(shortname)+7);
      strcpy(gt->fname,shortname);
      strcpy(buf,gt->fname);
      strcat(gt->fname,".gnut");
  } else {
    gt->fname=(char *) xmalloc(strlen(conf_get_str(dest_dir)) + strlen(qr->name)+250);
    gt->fname[0]=0;
    dpath=expand_path(conf_get_str(dest_dir));
    if (strlen(dpath)>1) {
      strcpy(gt->fname,dpath);
      if (dpath[strlen(dpath)-1]!=path_slash[0])
    strcat(gt->fname,path_slash);
    }
    strcat(gt->fname,shortname);
    strcpy(buf,gt->fname);
    strcat(gt->fname,".gnut");
    xfree(dpath);
  }
  
  // first we're going to see if the un-munged name exists, if so
  // then always make the download fail, because purportedly a good file
  // already exists
  g_debug(1,"checking if file \"%s\" exists\n",buf);
  
  ret=stat(buf,&ss);
  if (ret!=-1) {
    // woah! this file already exists, we shouldn't download it
    if (conf_get_int("verbose")) {
      printf("File \"%s\" already exists, download not started!\n",buf);
    }
    if (cache_download)
    cache_unlock();
    xfree(qr->name);
    xfree(qr);

    transfer_cleanup(gt,STATE_ERROR);

    return -1;
  }  
  
  g_debug(1,"testing file \"%s\" for size\n",gt->fname);
  
  gt->fsock=open(gt->fname,O_RDONLY);
  if (gt->fsock>=0) {
    g_debug(1,"file exists!\n");
    // the file exists, so we'll find out it's size...
    start=lseek(gt->fsock,0,SEEK_END);
    g_debug(1,"start=%i\n",start);
    if (start<0) start=0;
#ifdef WIN32
#undef close
    close(gt->fsock);
#define close(x) closesocket(x)
#else
    close(gt->fsock);
#endif
  }
  
  g_debug(3,"checked for existing, start=%i\n",start);
    
  // now that we're connected, we can send our request message....
  request=(char *) xmalloc(strlen(qr->name)+2049);
  if (start==0) {
    sprintf(request,"GET /get/%i/%s HTTP/1.0\r\nConnection: Keep-Alive\r\n\r\n",
      qr->ref, qr->name);
  } else {
    size=start;
    sprintf(request,"GET /get/%i/%s HTTP/1.0\r\nConnection: Keep-Alive\r\nRange: bytes=%i-\r\n\r\n",
      qr->ref, qr->name, start);
  }
  
  write(gt->sock,request,strlen(request));
  
  g_debug(3,"connect message sent\n");
  
  // now we look at the http response and make sure that the
  // correct number of newlines is returned....
  
  while (1) {
    ret=read_line(gt->sock,buf,sizeof(buf));
    if ((ret<=0) || (strstr(buf,"HTTP 4")!=NULL)) {
      if (cache_download) cache_unlock();
      xfree(request);
      xfree(qr->name);
      xfree(qr);

      transfer_cleanup(gt,STATE_ERROR);

      return -1;
    }
    if (strlen(buf)<=2) break;
  }
  
  g_debug(3,"read http headers...\n");
  
  gt->fsock=open(gt->fname,O_WRONLY|O_CREAT|O_APPEND,438);
  if (gt->fsock<0) {
    fprintf(stderr,"Couldn't open file %s for writing.  Transfer aborted\n",gt->fname);
  }
  
#ifdef HAVE_FLOCK
  ret=flock(gt->fsock,LOCK_EX | LOCK_NB);
#elif defined(HAVE_FCNTL)
  lock.l_type=F_WRLCK;
  lock.l_whence=SEEK_SET;
  lock.l_start=0;
  lock.l_len=0;
  ret=fcntl( gt->fsock, F_SETLK, &lock );
#else 
  ret=0;
#endif
  if (gt->fsock>0 && ret>=0) {
    gt->state=STATE_CONNECTED;        
    gt->bytes=start;
    count_downloads++;
    if (conf_get_int("verbose")>=1) {
      printf("\nDownload of %s started.\n",
        gt->fname);
    }

    if (gt->type==TYPE_RECV_PUSH) {
      // we need to remove this query struc
      g_debug(1,"removing query from list\n");
      query_push_remove(qr);
    }

    
    ret=gnut_transfer_loop(gt);
                // do not create incomplete cache files 
                if ((ret != 0) && cache_download) {
                  unlink(gt->fname);
                }
    if (ret==0 && conf_get_int("verbose")>=1) {
      printf("\nDownload of %s succeeded.\n",
        gt->fname);
      
      // since we  succeeded, let's change the name to what
      // it originally was intended to be
      if (strlen(gt->fname)>5) {
      
        c=&gt->fname[strlen(gt->fname)-5];
        g_debug(1,"c=%s\n",c);
        if (strncmp(c,".gnut",5)==0) {
          g_debug(1,"attempting to rename file...\n");
          c2=strdup(gt->fname);
          *c=0;
          rename(c2,gt->fname);
        }
      }
      
      if (gt->type==TYPE_RECV_PUSH) {
        g_debug(1,"finishing off query\n");
        // we succeeded, so we can finish off the query struc
        xfree(qr->name);
        xfree(qr);
      }
    } else {
      printf("\nDownload of %s failed after %i bytes.\n",
        gt->fname,gt->bytes);
      if (gt->type==TYPE_RECV_PUSH) {
        // we failed, so we should add the request back into
        // the list
        g_debug(1,"adding query back to list\n");
        query_push_add(qr);
      }
    }
    g_debug(4,"decreasing downloads\n");  
  }
  
  g_debug(3,"closing up transfer\n");
  xfree(request);

    if (cache_download) {
      cache_timestamp();
      rescan_cache();
      cache_unlock();
    }

  transfer_cleanup(gt,STATE_DEAD);
  
  g_debug(1,"returning %i\n",ret);
  return ret;
}

/* void *gnut_threads_outgoing(void *arg)
 *
 * this is the main thread for handling an outgoing connection
 * it does the connecting, negotiates with the remote servant
 * and cleans up after itself... arg is a pointer to a string
 * that indicates the host to connect to
 */
void *gnut_threads_outgoing(void *arg)
{
  gnut_connection *mgc;
  char *ptr;
  int ret;
  uchar cport[2];
  uint16 mport;

  g_debug(1,"entering arg=%s\n",arg);

  num_outgoing++;
  
  mgc=gnut_connection_new();
  
  mgc->ctype=TYPE_OUTGOING;
  mgc->cstate=STATE_CONNECTING;
  
  gnut_strdelimit(arg, ":", ' ');
  
  ptr=strchr(arg,' ');
  if (ptr!=NULL) {
    *(ptr++)=0;
    mgc->port=atoi(ptr);
  } else {
    mgc->port=6346;
  }

  mport=GUINT16_TO_LE(mgc->port);
  memcpy(cport,&mport,2);

  inet_aton(arg, (struct in_addr *) mgc->ip);

  mgc->sock=make_connection(arg, mgc->port, mgc->ip, 0);

  if  (mgc->sock<0) {
  /* printf("%s failed1\n", (char *) arg); %%%RPM */
    g_debug(1,"couldn't connect to: %s:%i\n",arg,mgc->port);
    mgc->cstate=STATE_CLEANABLE;

    host_remove(mgc->ip, cport);

    gnut_connection_delete(mgc);
    xfree(arg);
    num_outgoing--;
    return NULL;
  }

  mgc->cstate=STATE_NEGOTIATING;

  ret=send_welcome_message(mgc->sock, "0.4");

  if (ret<0) {
  /* printf("%s failed2\n", (char *) arg); %%%RPM */
    g_debug(1,"couldn't send message\n");
    close(mgc->sock);

    host_remove(mgc->ip,cport);

    gnut_connection_delete(mgc);
    g_debug(1,"free'ing arg\n");
    xfree(arg);
    g_debug(1,"done free'ing arg\n");
    num_outgoing--;
    pthread_exit(NULL);
    return NULL;
  }

  ret=read_welcome_message(mgc->sock, arg);

  if (ret==2) {
    char buf[256];
    // a password was requested, we'll try sending the
    // one in our config
    sprintf(buf,"%s\n",conf_get_str("password"));

    write(mgc->sock,buf,strlen(buf));

    // now we need to repeat the read for a welcome message
    ret=read_welcome_message(mgc->sock, arg);
    if (ret<0) {
    /* printf("%s failed3\n", (char *) arg); %%%RPM */
      printf("Incorrect password.\n");
    }
  }

  if (ret<0) {
    /* printf("%s failed4\n", (char *) arg); %%%RPM */
    g_debug(1,"couldn't read welcome message\n");
    close(mgc->sock);

    host_remove(mgc->ip,cport);
    gnut_connection_delete(mgc);
    g_debug(1,"free'ing arg\n");
    xfree(arg);
    g_debug(1,"done free'ing arg\n");
    num_outgoing--;
    pthread_exit(NULL);
    return NULL;
  }

  mgc->cstate=STATE_CONNECTED;

  ret = connection_loop(mgc);

  mgc->cstate=STATE_CLEANABLE;
  num_outgoing--;
  close(mgc->sock);
  host_remove(mgc->ip,cport);

  gnut_connection_delete(mgc);
  xfree(arg);

  pthread_exit(NULL);
  return NULL;
}
#ifdef HP1020
#undef PTHREADS_DRAFT4
#endif
#ifdef PTHREADS_DRAFT4
#include <pthread.h>
/*
 * Pthreads draft 4 implements a number of things differently, one of which
 * is that you can't statically initalise mutexes.  kxp@atl.hp.com
 */
 
void gnut_mutexes_init() {
  pthread_mutex_init(&gc_list_mutex,pthread_mutexattr_default);
  pthread_mutex_init(&query_packet_mutex,pthread_mutexattr_default);
  pthread_mutex_init(&send_mutex, pthread_mutexattr_default);
  pthread_mutex_init(&_g_debug_mutex,pthread_mutexattr_default);
  pthread_mutex_init(&make_connection_mutex, pthread_mutexattr_default);
  pthread_mutex_init(&transfer_mutex,pthread_mutexattr_default);
  pthread_mutex_init(&host_mutex,pthread_mutexattr_default);
  pthread_mutex_init(&monitor_mutex,pthread_mutexattr_default);
  pthread_mutex_init(&query_mutex, pthread_mutexattr_default);
  pthread_mutex_init(&push_mutex, pthread_mutexattr_default);
  pthread_mutex_init(&route_mutex, pthread_mutexattr_default);
  pthread_mutex_init(&share_mutex, pthread_mutexattr_default);
  pthread_mutex_init(&cache_mutex, pthread_mutexattr_default);
}
#endif

